type point = float * float;;

let dist ((x1,y1), (x2,y2)) =
  sqrt ((x1-.x2)*.(x1-.x2)+.(y1-.y2)*.(y1-.y2));;

let rec filter f = function
  | [] -> []
  | t::q -> let q = filter f q in
            if f t then t::q
            else q;;

let rec strip xmin xmax =
  filter (fun (x,y) -> xmin <= x && x <= xmax);;

let rec closest_neighbour (p : point) (acc : point * point) (d : float) (n : int) (l : point list) =
  match l, n with
  | [], _ | _, 0 -> acc, d
  | p2::q, n -> let d2 = dist (p,p2) in
             if d2 < d then closest_neighbour p (p, p2) d2 (n-1) q
             else closest_neighbour p acc d (n-1) q;;

let rec closest_in_strip (acc : point * point) (d : float) (l : point list) =
  match l with
  | [] -> acc
  | t::q -> let acc, d = closest_neighbour t acc d 7 q in
            closest_in_strip acc d q;;

let rec len = function
  | [] -> 0
  | t::q -> 1 + len q;;

let rec select n l = match l, n with
  | [], _ -> [], []
  | l, 0 -> [], l
  | t::q, n -> let (l1,l2) = select (n-1) q in
               t::l1, l2;;

let halves l = select (len l / 2) l;;

let rec closest px py = match px with
  | [] | [_] -> raise (Invalid_argument "Not enough points")
  | [p1;p2] -> p1, p2
  | [p1;p2;p3] -> let d1 = dist (p2,p3) in
                  let d2 = dist (p1,p3) in
                  let d3 = dist (p1,p2) in
                  if d1 <= d2 && d1 <= d3 then (p2,p3)
                  else if d2 <= d3 then (p1,p3)
                  else (p1,p2)
  | px -> let (px1,px2) = halves px in
          let best1 = closest px1 py in
          let best2 = closest px2 py in
          let d1, d2 = dist best1, dist best2 in
          let best, d = if d1 <= d2 then best1, d1 else best2, d2 in
          match px2 with
          | (xmed, _)::_ ->
             let s = strip (xmed -. d) (xmed +. d) py in
             closest_in_strip best d s
          | _ -> assert false
;;

(*#require "graphics";;*)
open Graphics;;

let points_x = ref [];;
let points_y = ref [];;

let rec add_point_x (x,y) = function
  | [] -> [(x,y)]
  | t::q -> if (x,y) < t then (x,y)::t::q
            else t::add_point_x (x,y) q;;

let rec add_point_y (x,y) = function
  | [] -> [(x,y)]
  | (a,b)::q -> if (y,x) < (b,a) then (x,y)::(a,b)::q
                else (a,b)::add_point_y (x,y) q;;

let draw_point (x,y) =
  set_color blue;
  fill_circle (int_of_float x) (int_of_float y) 3;
  set_color black;;

let get_coord () =
  let st = wait_next_event [Button_down] in
  float_of_int st.mouse_x, float_of_int st.mouse_y;;

open_graph "";;
let run () =
  while true do
    let (x,y) = get_coord () in
    remember_mode true;
    points_x := add_point_x (x,y) !points_x;
    points_y := add_point_y (x,y) !points_y;
    draw_point (x,y);
    remember_mode false;
    synchronize ();
    try let ((x1,y1),(x2,y2)) = closest !points_x !points_y in
        moveto (int_of_float x1) (int_of_float y1);
        lineto (int_of_float x2) (int_of_float y2);
    with Invalid_argument s -> ();
  done;;

let () = run ();;
